/*
Copyright (C) 2005         linap

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA 02111-1307, USA.
*/

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

import javax.imageio.ImageIO;
import leveledit.COLOR;

/**
 * Bitmap zu snEADy-Map-Konverter
 * 
 * @author linap, schnueptus
 */
public class LevelEditor {
	private BufferedImage srcImage;
	private int width, height;
	private final boolean debug;
	
	public LevelEditor( boolean d ) {
		debug = d;
	}
	
	private boolean isNotValid(int x, int y) {
		return ( x < 0 || x >= width || y < 0 || y >= height );
	}
	
	private boolean isWand(int x, int y) {
		if( isNotValid(x,y) ) return false;
		return srcImage.getRGB(x,y) == COLOR.BLACK || srcImage.getRGB(x,y) == COLOR.GREEN;
	}
	
	private boolean setWand(int x, int y) {
		if( x < 0 || x >= width || y < 0 || y >= height ) return false;
		if( srcImage.getRGB(x,y) == COLOR.BLACK || srcImage.getRGB(x,y) == COLOR.GREEN ) return false;
		srcImage.setRGB( x, y, COLOR.GREEN );
		return true;
	}
		
	private int getWallNeighbors( int x, int y ) {
		int nb = 0;
		if( isWand(x+1,y) ) nb++;
		if( isWand(x-1,y) ) nb++;
		if( isWand(x,y+1) ) nb++;
		if( isWand(x,y-1) ) nb++;
		return nb;
	}
	
	private void paintSuitableNeighbor(int x, int y,int mask) {
		int which = 0;
		int maxNb = 0;
		int nb;
		
		if( (mask & 0x01) != 0 ) {				
			nb = getWallNeighbors(x,y+1);
			if( !isNotValid(x,y+1) && !isWand(x,y+1) && nb > maxNb ) {
				maxNb = nb;
				which = 1;
			}
			
			nb = getWallNeighbors(x,y-1);
			if( !isNotValid(x,y-1) && !isWand(x,y-1) && nb > maxNb ) {
				maxNb = nb;
				which = 3;
			}
		} 
		if( (mask & 0x10) != 0 ) {		
			nb = getWallNeighbors(x+1,y);
			if( !isNotValid(x+1,y) && !isWand(x+1,y) && nb > maxNb ) {
				maxNb = nb;
				which = 0;
			}
			
			nb = getWallNeighbors(x-1,y);
			if( !isNotValid(x-1,y) && !isWand(x-1,y) && nb > maxNb ) {
				maxNb = nb;
				which = 2;
			}
		}			
		
		switch( which ) {
		case 0:	setWand( x+1, y ); break;
		case 1:	setWand( x, y+1 ); break;
		case 2:	setWand( x-1, y ); break;
		case 3:	setWand( x, y-1 ); break;
		}
	}
	
	private boolean scanForLars(int lev) {
		int width = srcImage.getWidth();
		int height = srcImage.getHeight();
		boolean cor = false;
		for( int y = 0; y<height; y++ ) {
			for( int x = 0; x<width; x++ ) {
				int nb = getWallNeighbors( x, y );
				if( !isWand(x,y) ) continue;
				if( nb == 0 ) {
					if( debug )
						System.out.println( "malformed map: nb=0 @ " + x + ";" + y );
					srcImage.setRGB(x,y,COLOR.RED);
					cor = true;
					continue;
				}
				if( nb == 1 ) {
					if( debug )
						System.out.println( "malformed map: nb=1 @ " + x + ";" + y );
					if( lev > 0 ) {
						paintSuitableNeighbor(x,y,3);
					}
					cor = true;
					continue;
				}
				if( nb == 2 ) {
					if( isWand(x+1,y) && isWand(x-1,y) ) {
						if( debug )
							System.out.println( "malformed map: nb=2 @ " + x + ";" + y );
						if( lev > 0 ) {
							paintSuitableNeighbor(x,y,1);
						}
						cor = true;
					} else if( isWand(x,y+1) && isWand(x,y-1) ) {
						if( debug )
							System.out.println( "malformed map: nb=2 @ " + x + ";" + y );
						if( lev > 0 ) {
							paintSuitableNeighbor(x,y,2);
						}
						cor = true;
					}					
					continue;
				}
				if( nb >= 3 && 
						( !(isWand(x+1,y+1)||isWand(x-1,y-1)) 
								|| !(isWand(x-1,y+1)||isWand(x+1,y-1)) ) ) {
					
					if( debug )
						System.out.println( "malformed map: cf. Lars @ " + x + ";" + y );
					if( !(isWand(x+1,y+1)||isWand(x-1,y-1)) ) {
						if( !isNotValid(x+1,y+1) ) {
							srcImage.setRGB(x+1,y+1, COLOR.GREEN);
						} else {
							srcImage.setRGB(x-1,y-1, COLOR.GREEN);
						}
					}
					if( !(isWand(x-1,y+1)||isWand(x+1,y-1)) ) {
						if( !isNotValid(x-1,y+1) ) {
							srcImage.setRGB(x-1,y+1, COLOR.GREEN);
						} else {
							srcImage.setRGB(x+1,y-1, COLOR.GREEN);
						}
					}
					cor = true;
				}
			}
		}
		
		return cor;
	}
	/**
	 * png2txt funktion!
	 * @param inFilename
	 * @param outFilename
	 * @throws IOException
	 */
	public void parse( String inFilename, String outFilename ) throws IOException {
		File f = new File( inFilename );		
		srcImage = ImageIO.read( f );
		width = srcImage.getWidth();
		height = srcImage.getHeight();
					
		int i;
		for( i=0; scanForLars(i); i++);
		if( i > 1 ) {
			System.err.println( " Die Karte entspricht nicht dem snEADy-Standard !!! ");
			System.err.println( " Korrekturvorschlaege befinden sich in der Datei: " + 					
					"try_"+inFilename);
			System.err.println( "\n Es wurde _keine_ snEADy-Karte erstellt. ");
			ImageIO.write( srcImage, "png", new File( "try_"+inFilename ) );
			return;
		}
		/**
		 * TODO: sind die exits 5x5 felder?
		 */
		
		
		FileWriter fw = new FileWriter( outFilename );
		fw.write( "<name> " + inFilename.substring(0,inFilename.length()-4) + "\n" );
		fw.write( "<size> " + height + "x" + width + "\n" );
		fw.write( "<maxPlayer> 10\n" );
		/**
		 * TODO: anzahl der startpunkte + keiner doppelt!
		 */
		fw.write( "<author> unknown\n" );
		fw.write( "<comment>created with linap's png2smp-converter\n<endcomment>\n" );
		
		fw.write( "<playField>\n" );		
		for( int y = 0; y<height; y++ ) {
			for( int x = 0; x<width; x++ ) {
				int c = srcImage.getRGB( x, y );
				if( c == COLOR.BLACK ) {
					fw.write( "#" );
				} 
				else if( c == COLOR.RED ) {
					fw.write( "A" );
				} 
				else if( c == COLOR.SNAKE0 ) {fw.write( "0" );}
				else if( c == COLOR.SNAKE1 ) {fw.write( "1" );} 
				else if( c == COLOR.SNAKE2 ) {fw.write( "2" );} 
				else if( c == COLOR.SNAKE3 ) {fw.write( "3" );} 
				else if( c == COLOR.SNAKE4 ) {fw.write( "4" );} 
				else if( c == COLOR.SNAKE5 ) {fw.write( "5" );} 
				else if( c == COLOR.SNAKE6 ) {fw.write( "6" );} 
				else if( c == COLOR.SNAKE7 ) {fw.write( "7" );} 
				else if( c == COLOR.SNAKE8 ) {fw.write( "8" );} 
				else if( c == COLOR.SNAKE9 ) {fw.write( "9" );} 
				else if( c == COLOR.GOODIE1 ) {fw.write( "a" );}
				else if( c == COLOR.GOODIE2 ) {fw.write( "b" );} 
				else if( c == COLOR.GOODIE3 ) {fw.write( "c" );} 
				else if( c == COLOR.GOODIE4 ) {fw.write( "d" );} 
				else if( c == COLOR.GOODIE5 ) {fw.write( "e" );} 
				else {
					fw.write( "." );
				}
			}
			fw.write( "\n" );
		}
		fw.write( "<endPlayfield>\n" );
		
		fw.close();
	}
	
	private void colorSquare(int x, int y, int a, int c ) {
		for( int px = x; px<x+a; px++ ) {
			for( int py = y; py<y+a; py++ ) {
				srcImage.setRGB( px, py, c );
			}
		}
	}
	
	/**
	 * txt2png function!
	 * @param inFilename
	 * @param outFilename
	 * @param factor
	 * @throws IOException
	 */
	public void parseLevel( String inFilename, String outFilename, int factor ) throws IOException {
		File f = new File( outFilename );
		
		//default colors
		int COLOR0 = COLOR.SNAKE0;
		int COLOR1 = COLOR.SNAKE1;
		int COLOR2 = COLOR.SNAKE2;
		int COLOR3 = COLOR.SNAKE3;
		int COLOR4 = COLOR.SNAKE4;
		int COLOR5 = COLOR.SNAKE5;
		int COLOR6 = COLOR.SNAKE6;
		int COLOR7 = COLOR.SNAKE7;
		int COLOR8 = COLOR.SNAKE8;
		int COLOR9 = COLOR.SNAKE9;
		
		int GOODIE1 = COLOR.GOODIE1;
		int GOODIE2 = COLOR.GOODIE2;
		int GOODIE3 = COLOR.GOODIE3;
		int GOODIE4 = COLOR.GOODIE4;
		int GOODIE5 = COLOR.GOODIE5;
		
		/**
		 * TODO: may to load a other color configuration!
		 */
		
		BufferedReader fr = new BufferedReader( new FileReader( inFilename ) );
		String line;
		boolean field = false;
		srcImage = null;
		LinkedList l = new LinkedList();
		int width = 0;
		while( (line = fr.readLine()) != null ) {
			if( line.startsWith("<playField>") ) field = true;
			if( line.startsWith("<endPlayfield>") ) field = false;
			
			if( field ) {				
				l.add( line );
				width = line.length();
			}
		}
		l.remove( 0 );
		fr.close();
		
		srcImage = new BufferedImage( factor*width, factor*l.size(), BufferedImage.TYPE_INT_ARGB );
		int x=0, y=0;
		for( Iterator i=l.iterator(); i.hasNext(); ) {
			line = (String) i.next();			
			for( char ch=line.charAt(x); x < width; ) {			
				if( ch == '#' ) colorSquare( factor*x, factor*y, factor, COLOR.BLACK );
				else if( ch == 'A' ) colorSquare( factor*x, factor*y, factor, COLOR.RED );
				else if( ch == '0' ) colorSquare( factor*x, factor*y, factor, COLOR0 );
				else if( ch == '1' ) colorSquare( factor*x, factor*y, factor, COLOR1 );
				else if( ch == '2' ) colorSquare( factor*x, factor*y, factor, COLOR2 );
				else if( ch == '3' ) colorSquare( factor*x, factor*y, factor, COLOR3 );
				else if( ch == '4' ) colorSquare( factor*x, factor*y, factor, COLOR4 );
				else if( ch == '5' ) colorSquare( factor*x, factor*y, factor, COLOR5 );
				else if( ch == '6' ) colorSquare( factor*x, factor*y, factor, COLOR6 );
				else if( ch == '7' ) colorSquare( factor*x, factor*y, factor, COLOR7 );
				else if( ch == '8' ) colorSquare( factor*x, factor*y, factor, COLOR8 );
				else if( ch == '9' ) colorSquare( factor*x, factor*y, factor, COLOR9 );
				else if( ch == 'a' ) colorSquare( factor*x, factor*y, factor, GOODIE1 );
				else if( ch == 'b' ) colorSquare( factor*x, factor*y, factor, GOODIE2 );
				else if( ch == 'c' ) colorSquare( factor*x, factor*y, factor, GOODIE3 );
				else if( ch == 'd' ) colorSquare( factor*x, factor*y, factor, GOODIE4 );
				else if( ch == 'e' ) colorSquare( factor*x, factor*y, factor, GOODIE5 );
				else colorSquare( factor*x, factor*y, factor, COLOR.WHITE );
				x++;
				if( x < width ) ch=line.charAt(x);
			}			
			x=0;
			y++;
		}
		
		ImageIO.write( srcImage, "png", f);		
		
	}
	
	public static void usage() {
		System.out.println("Bitmap zu snEADy-Map-Konverter");
		System.out.println("  Autor: linap & Schnueptus (2005) - veroeffentlich unter der GNU Public Licence\n");
		System.out.println("Aufruf: java LevelEditor <s|p[x]> <eingangsfile> <ausgangsfile>");
		System.out.println("Modi:");
		System.out.println("\ts\tliest ein PNG ein und gibt eine snEADy-Map aus");
		System.out.println("\tp\tliest eine snEADy-Map ein und gibt sie als PNG aus");
		System.out.println("\t[x]\twenn hinter p eine Zahl x angegeben ist, wird die PNG-Grafik");
		System.out.println("\t\tum diesen Faktor vergroessert");
		System.out.println();
		System.out.println("<ausgangsfile> enthaelt das Ausgangsbitmap bzw. die Map");
		System.out.println("<eingangsfile> enthaelt anschliessend die Map oder das PNG");
	}
	
	public static void main(String[] args) {
		if( args.length != 3 ) {
			usage();
			System.exit( -1 );
		}
		char modus = args[0].charAt(0);
		if( modus == 's' ) {
			try {
				String pngFile = args[1];
				String smpFile = args[2];
				LevelEditor n = new LevelEditor( false );
				n.parse( pngFile, smpFile );
			} catch( IOException e ) {
				e.printStackTrace();
			}
		} else if( modus == 'p' ) {
			try {				
				String smpFile = args[1];
				String pngFile = args[2];
				LevelEditor n = new LevelEditor( false );
				int factor = 1;
				try {
					factor = Integer.parseInt( args[0].substring(1) );
				} catch ( Exception ex ) {
					factor = 1;
				}
				n.parseLevel( smpFile, pngFile, factor );
			} catch( IOException e ) {
				e.printStackTrace();
			}
		} else {
			usage();
		}
		
	}
}
